<?php

declare(strict_types=1);

namespace Drupal\Tests\otlog\Unit;

use Drupal\Core\Logger\LogMessageParser;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\Core\Utility\Error;
use Drupal\Tests\UnitTestCase;
use Drupal\Tests\otlog\Constraint\LogRecordConstraint;
use Drupal\otlog\Logger\OtLog;
use OpenTelemetry\API\Logs\LoggerInterface;
use OpenTelemetry\API\Logs\LoggerProviderInterface;
use OpenTelemetry\API\Logs\Severity;
use OpenTelemetry\SemConv\TraceAttributes;
use PHPUnit\Framework\Attributes\CoversClass;
use PHPUnit\Framework\Attributes\Group;
use PHPUnit\Framework\MockObject\MockObject;
use Psr\Log\LogLevel;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Tests OpenTelemetry log bridge.
 */
#[Group('otlog')]
#[CoversClass(OtLog::class)]
final class OtLogTest extends UnitTestCase {

  /**
   * Test that nothing happens if logger is not enabled.
   */
  public function testLoggerDisabled(): void {
    $logger = $this->logger(enabled: FALSE);
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('test');
    $channel->addLogger($otlog);
    $channel->log(LogLevel::INFO, 'This message is skipped');
  }

  /**
   * Test that debug logs are emitted with the correct channel and severity.
   */
  public function testLogDebugLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[cron] [debug] A worker for locale_translation queue suspended further processing of the queue.',
        severityText: 'debug',
        severityNumber: Severity::DEBUG->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('cron');
    $channel->addLogger($otlog);
    $channel->debug('A worker for @queue queue suspended further processing of the queue.', [
      '@queue' => 'locale_translation',
    ]);
  }

  /**
   * Test that info logs are emitted with the correct channel and severity.
   */
  public function testLogInfoLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[taxonomy] [info] Created new term PHP.',
        severityText: LogLevel::INFO,
        severityNumber: Severity::INFO->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('taxonomy');
    $channel->addLogger($otlog);
    $channel->info('Created new term %term.', ['%term' => 'PHP']);
  }

  /**
   * Test that notice logs are emitted with the correct channel and severity.
   */
  public function testLogNoticeLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[cron] [notice] Cron could not run because an invalid key was used.',
        severityText: LogLevel::NOTICE,
        severityNumber: Severity::INFO2->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('cron');
    $channel->addLogger($otlog);
    $channel->notice('Cron could not run because an invalid key was used.');
  }

  /**
   * Test that warning logs are emitted with the correct channel and severity.
   */
  public function testLogWarningLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[locale] [warning] No configuration objects have been updated.',
        severityText: LogLevel::WARNING,
        severityNumber: Severity::WARN->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('locale');
    $channel->addLogger($otlog);
    $channel->warning('No configuration objects have been updated.');
  }

  /**
   * Test that error logs are emitted with the correct channel and severity.
   */
  public function testLogErrorLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[announcements_feed] [error] The feed format is not valid.',
        severityText: LogLevel::ERROR,
        severityNumber: Severity::ERROR->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('announcements_feed');
    $channel->addLogger($otlog);
    $channel->error('The feed format is not valid.');
  }

  /**
   * Test that critical logs are emitted with the correct channel and severity.
   */
  public function testLogCriticalLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[entity_reference] [critical] The tools entity reference field (entity_type: tool, bundle: chainsaw) no longer has any valid bundle it can reference.',
        severityText: LogLevel::CRITICAL,
        severityNumber: Severity::ERROR2->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('entity_reference');
    $channel->addLogger($otlog);
    $channel->critical('The %field_name entity reference field (entity_type: %entity_type, bundle: %bundle) no longer has any valid bundle it can reference.', [
      '%field_name' => 'tools',
      '%entity_type' => 'tool',
      '%bundle' => 'chainsaw',
    ]);
  }

  /**
   * Test that alert logs are emitted with the correct channel and severity.
   */
  public function testLogAlertLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[filter] [alert] Missing filter plugin: markdown.',
        severityText: LogLevel::ALERT,
        severityNumber: Severity::ERROR3->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('filter');
    $channel->addLogger($otlog);
    $channel->alert('Missing filter plugin: %filter.', ['%filter' => 'markdown']);
  }

  /**
   * Test that emergency logs are emitted with the correct channel and severity.
   */
  public function testLogEmergencyLevel(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[album] [emergency] on planet earth!',
        severityText: LogLevel::EMERGENCY,
        severityNumber: Severity::FATAL->value,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('album');
    $channel->addLogger($otlog);
    $channel->emergency('on planet earth!');
  }

  /**
   * Test that backtrace is extracted to an attribute.
   */
  public function testLogBacktrace(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[php] [error] A message with a trace',
        attributes: [
          TraceAttributes::CODE_STACKTRACE => "stack trace\nanother frame\n",
        ],
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('php');
    $channel->addLogger($otlog);
    $channel->error('A message with a trace @backtrace_string', [
      '@backtrace_string' => "stack trace\nanother frame\n",
    ]);
  }

  /**
   * Test that error details are extracted into attributes.
   */
  public function testErrorDetails(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        body: '[php] [error] DbError: No such field "score" in DB->select() (line 42 of /path/to/db.php).',
        attributes: [
          TraceAttributes::ERROR_TYPE => 'DbError',
          TraceAttributes::CODE_FUNCTION => 'DB->select()',
          TraceAttributes::CODE_FILEPATH => '/path/to/db.php',
          TraceAttributes::CODE_LINENO => 42,
        ],
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('php');
    $channel->addLogger($otlog);
    $channel->error(Error::DEFAULT_ERROR_MESSAGE, [
      '@message' => 'No such field "score"',
      '%type' => 'DbError',
      '%function' => 'DB->select()',
      '%file' => '/path/to/db.php',
      '%line' => 42,
    ]);
  }

  /**
   * Test that timestamp is handled correctly.
   */
  public function testTimestamp(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        timestamp: 5555074026000000000,
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('time');
    $channel->addLogger($otlog);
    $channel->info('', ['timestamp' => 5555074026]);
  }

  /**
   * Test that current user is recorded in the user.id attribute.
   */
  public function testCurrentUser(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        attributes: [TraceAttributes::USER_ID => 543210],
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $account = $this->createMock(AccountInterface::class);
    $account->expects($this->once())
      ->method('id')
      ->willReturn(543210);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('user');
    assert($account instanceof AccountInterface);
    $channel->setCurrentUser($account);
    $channel->addLogger($otlog);
    $channel->info('Account updated');
  }

  /**
   * Test that request uri is recorded in url.full attribute.
   */
  public function testRequestUri(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        attributes: [TraceAttributes::URL_FULL => 'http://localhost/node/1/edit'],
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $request = Request::create('/node/1/edit');

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('content', $request);
    $channel->addLogger($otlog);
    $channel->info('Node updated');
  }

  /**
   * Test that ip address is recorded in client.address attribute.
   */
  public function testClientIp(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        attributes: [TraceAttributes::CLIENT_ADDRESS => '2001:db8::fee'],
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $request = Request::create('/', server: [
      'REMOTE_ADDR' => '2001:db8::fee',
    ]);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('flood', $request);
    $channel->addLogger($otlog);
    $channel->info('');
  }

  /**
   * Test that legacy ip address is recorded in client.address attribute.
   */
  public function testClientLegacyIp(): void {
    $logger = $this->logger(enabled: TRUE);
    $logger->expects($this->once())
      ->method('emit')
      ->with(new LogRecordConstraint(
        attributes: [TraceAttributes::CLIENT_ADDRESS => '203.0.113.31'],
      ));
    $loggerProvider = $this->loggerProvider($logger);

    $request = Request::create('/', server: [
      'REMOTE_ADDR' => '203.0.113.31',
    ]);

    $otlog = new OtLog(new LogMessageParser(), $loggerProvider);
    $channel = $this->loggerChannel('flood', $request);
    $channel->addLogger($otlog);
    $channel->info('');
  }

  /**
   * Returns a logger channel populated with the given request.
   */
  protected function loggerChannel(string $name, ?Request $request = NULL): LoggerChannelInterface {
    $requestStack = new RequestStack();
    $requestStack->push($request ?? Request::create('/'));

    $channel = new LoggerChannel($name);
    $channel->setRequestStack($requestStack);

    return $channel;
  }

  /**
   * Creates a logger provider mock returning the given logger.
   */
  protected function loggerProvider(LoggerInterface $logger): LoggerProviderInterface|MockObject {
    $provider = $this->createMock(LoggerProviderInterface::class);
    $provider->expects($this->once())
      ->method('getLogger')
      ->with('org.drupal.otlog')
      ->willReturn($logger);
    return $provider;
  }

  /**
   * Creates a logger mock with an expectation for the isEnabled method.
   */
  protected function logger(bool $enabled): LoggerInterface|MockObject {
    $logger = $this->createMock(LoggerInterface::class);
    $logger->expects($this->once())
      ->method('isEnabled')
      ->willReturn($enabled);
    return $logger;
  }

}
